/* 3D Bouncing Ball by Brad Whitsitt 10/3/90 version 1.0 */


#include <conio.h>
#include <graph.h>
#include <math.h>
#include <stdio.h>
#include <ctype.h>

typedef struct                    /* define structure for ball info */
{
   double x;
   double y;
   double z;
   double velx;
   double vely;
   double velz;
   double newx;
   double newy;
   double newz;
   double newvelx;
   double newvely;
   double newvelz;
   double accelx;
   double accely;
   double accelz;
   double friction;
   double xscreen;
   double yscreen;
}attributes;                             /* call it attributes */

attributes ball[3];                      /* define array of 3 structures for ball info */
attributes start[3];                     /* define array of 3 structures for start info for each ball */


int screendist=8000;         /* distance from screen to beginning of world in pixels */
int view=400;                /* view distance from your eye to screen in pixels */
int xview=3000;              /* point on world at which view is taken for x axis */
int yview=10000;             /* point on world at which view is taken for y axis */
int xoffset=320;             /* slide the world to the right by this number of pixels */
int yoffset=240;             /* slide the world down by this number of pixels */
int number=1;
int video;                /* video mode store */
int maxx;                 /* screen size */
int maxy;
int xworld=20000;         /* set size of world in x,y and z */
int yworld=20000;
int zworld=20000;



char mainmenu(void);                        /* set up main menu */
void instructions(void);                    /* tutorial */
double data_convert(char in[15]);           /* data conversion for input of ball data */
void project_world(void);                   /* create 3D view of world */
void ballsize(int num);                     /* make ball size larger near the front */
void point_convert(int num);                /* convert x y and z to screen coordinates */
void initialize(void);                      /* get new data from operator */
void set_default(void);                     /* set default ball data */
void setscreen(void);                       /* put in graphics and set border */
void reflectxlower(int num);         /* calcultate bounce left wall */
void reflectxupper(int num);         /* calcultate bounce right wall */
void reflectylower(int num);         /* calcultate bounce bottom */
void reflectyupper(int num);         /* calcultate bounce top */
void reflectzlower(int num);         /* calcultate bounce front */
void reflectzupper(int num);         /* calcultate bounce back */
void movexyz(int num);                             /* move ball to next point */
void checkboundary(int num);                       /* check for bounce */
void slow(int num);                                /* friction effect */
void beep(unsigned duration, unsigned frequency);  /* noise for bounce */
int graphics_mode(void);                           /* check graphics stuff out */
void init_instructions(void);                      /* set world position for instruction screen */
void init_bounce(void);                            /* set world position for bounce screen */



main()
{
   int num;                           /* ball number for calculations */
   char selection;                    /* desired selection from main menu */

   video=graphics_mode();             /* get video mode for later use */
   set_default();
   do
   {
   switch(selection=mainmenu())       /* main menu loop */
      {
      case '1':
	 instructions();
	 break;
      case '2':
	 initialize();
	 break;
      case '3':
	 init_bounce();
	 setscreen();
	 _setbkcolor(_BLUE);          /* background blue */
	 _setcolor(_BRIGHTWHITE);     /* world boundary brightwhite */
	 project_world();
	 _setcolor(14);               /* make ball yellow */
	 ball[0]=start[0];            /* set start values for bouncing */
	 ball[1]=start[1];
	 ball[2]=start[2];

	 do
	 {
	    num=number-1;
	    for(;num > -1;num--)      /* loop for number of balls */
	    {
	       movexyz(num);          /* move ball */
	       slow(num);             /* friction effect */
	       checkboundary(num);    /* look for bounce */
	       point_convert(num);    /* convert xyz to screen point */
	       _setpixel(ball[num].xscreen,ball[num].yscreen);
	       ballsize(num);         /* adjust ball size to improve depth effect */

	    }
	 }while(!kbhit());
	 _setvideomode(_DEFAULTMODE);
	 break;
	 case '4':
	 printf("bye!!\n");exit(0);   /* exit program */
	 break;
      }
   } while(selection!=4);             /* main menu loop */

}

void init_instructions(void)
{
   xworld=7000;
   yworld=7000;
   zworld=7000;
   xview=1500;            /* point on front wall where view will be from */
   yview=1500;
   xoffset=320;
   yoffset=240;
   screendist=2000;       /* how far is eye from screen */
   view=280;

}

void init_bounce(void)
{
   float temp1;
   float temp2;

   xworld=20000;            /* rest world size stuff */
   yworld=20000;
   zworld=20000;
   screendist=8000;         /* distance from screen to beginning of world in pixels */
   view=370;                /* view distance from your eye to screen in pixels */
   xview=4000;              /* point on world at which view is taken for x axis */
   yview=12000;             /* point on world at which view is taken for y axis */


   temp1=(float)xview;
   temp2=(float)xworld;
   xoffset=(temp1/temp2*640)+80;        /* slide the world to the right by this number of pixels */
   temp1=yview;
   temp2=yworld;
   yoffset=temp1/temp2 * 480;           /* slide the world down by this number of pixels */

}

void ballsize(int num)
{
   if(ball[num].z<9000)_setpixel(ball[num].xscreen+1,ball[num].yscreen);
   if(ball[num].z<7000)_setpixel(ball[num].xscreen,ball[num].yscreen+1);
   if(ball[num].z<5000)_setpixel(ball[num].xscreen+1,ball[num].yscreen+1);
   if(ball[num].z<2000)_setpixel(ball[num].xscreen-1,ball[num].yscreen);
   if(ball[num].z<1000)_setpixel(ball[num].xscreen-1,ball[num].yscreen+1);
}

void checkboundary(int num)
{
   if(ball[num].newx > xworld) reflectxupper(num);
   if(ball[num].newx < 1) reflectxlower(num);
   if(ball[num].newy > yworld) reflectyupper(num);
   if(ball[num].newy < 1) reflectylower(num);
   if(ball[num].newz > yworld) reflectzupper(num);
   if(ball[num].newz < 1) reflectzlower(num);
   ball[num].x=ball[num].newx;             /* modify position */
   ball[num].velx=ball[num].newvelx;       /* modify velocity */
   ball[num].y=ball[num].newy;                    /* y */
   ball[num].vely=ball[num].newvely;
   ball[num].z=ball[num].newz;                    /* z */
   ball[num].velz=ball[num].newvelz;

}

void instructions(void)                /* print tutorial to screen */
{

   init_instructions();
   setscreen();
   _setbkcolor(_BLUE);             /* set background color */
   _setcolor(_BRIGHTWHITE);        /* set world boundary color to brightwhite */
   project_world();

   _settextcolor(14);              /* set text color to yellow */
   _settextposition(2,33);
   _outtext("INSTRUCTIONS");
   _settextposition(3,5);
   _outtext("ORIGIN OF THE WORLD");
   _settextposition(4,5);
   _outtext("X=0,Y=0,Z=0");
   getch();
   _settextposition(6,2);
   _outtext("POSITIVE ACCELERATION");
   _settextposition(7,2);
   _outtext("AND VELOCITY FOR:");
   _settextposition(8,2);
   _outtext("X IS TO THE RIGHT");
   _settextposition(9,2);
   _outtext("Y IS DOWN");
   _settextposition(10,2);
   _outtext("Z IS TO THE BACK");
   _settextposition(12,7);
   getch();
   _outtext("THE SIZE OF THE WORLD:");
   _settextposition(13,7);
   _outtext("20000 x 20000 x 20000 PIXELS");
   getch();
   _settextposition(16,30);
   _outtext("THE FRONT WALL IS MADE OF GLASS");
   getch();
   _setcolor(2);                  /* make ball greeen */
   _setpixel(-150,20);            /* set ball */
   _setpixel(-151,20);
   _setpixel(-150,21);
   _setpixel(-151,21);
   _setcolor(4);                  /* make the pointer red */
   _moveto(-145,21);
   _lineto(-135,17);              /* top of pointer */
   _moveto(-145,21);
   _lineto(-138,28);              /* bottom of pointer */
   _moveto(-145,21);
   _lineto(90,60);                /* line to ball message */
   _settextposition(19,28);
   _outtext("THE BALL IS MADE OF 100% ELASTIC VGA RUBBER");
   getch();
   _settextposition(24,14);
   _outtext("HAVE FUN EXPERIMENTING IN THE WORLD OF THE BOUNCING BALL!");
   getch();
   _setvideomode(_DEFAULTMODE);
}

void initialize()                      /* collect data from operator */
{
   int i;
   char in[15];
   char text[80];

   _setbkcolor(0L);
   _clearscreen(_GCLEARSCREEN);
   _settextcolor(9);                   /* set text color to blue */
   _outtext("                        ** SET START PARAMETERS **\n");

   _outtext("HOW MANY BALLS? (3 MAX):");
   scanf("%i",&number);
   if(number>3)number=3;


   for(i=0;i<number;i++)
   {
      _outtext("\n\n");
      sprintf(text,"**** ENTER THE PARAMETERS FOR BALL %i ****",(i+1));
      _outtext(text);

      _outtext("\n");
      sprintf(text,"X POSITION %0.0lf - ",start[i].x);
      _outtext(text);
      fflush( stdin );
      gets(in);
      if(!(iscntrl(in[0])))start[i].x=data_convert(in);

      _outtext("\n");
      sprintf(text,"Y POSTIION %0.0lf - ",start[i].y);
      _outtext(text);
      fflush( stdin );
      gets(in);
      if(!(iscntrl(in[0])))start[i].y=data_convert(in);

      _outtext("\n");
      sprintf(text,"Z POSTIION %0.0lf - ",start[i].z);
      _outtext(text);
      fflush( stdin );
      gets(in);
      if(!(iscntrl(in[0])))start[i].z=data_convert(in);

      _outtext("\n");
      sprintf(text,"X VELOCITY %0.2lf - ",start[i].velx);
      _outtext(text);
      fflush( stdin );
      gets(in);
      if(!(iscntrl(in[0])))start[i].velx=data_convert(in);

      _outtext("\n");
      sprintf(text,"Y VELOCITY %0.02f - ",start[i].vely);
      _outtext(text);
      fflush( stdin );
      gets(in);
      if(!(iscntrl(in[0])))start[i].vely=data_convert(in);

      _outtext("\n");
      sprintf(text,"Z VELOCITY %0.02f - ",start[i].velz);
      _outtext(text);
      fflush( stdin );
      gets(in);
      if(!(iscntrl(in[0])))start[i].velz=data_convert(in);

      _outtext("\n");
      sprintf(text,"X ACCELERATION %0.3lf - ",start[i].accelx);
      _outtext(text);
      fflush( stdin );
      gets(in);
      if(!(iscntrl(in[0])))start[i].accelx=data_convert(in);

      _outtext("\n");
      sprintf(text,"Y ACCELERATION %0.3lf - ",start[i].accely);
      _outtext(text);
      fflush( stdin );
      gets(in);
      if(!(iscntrl(in[0])))start[i].accely=data_convert(in);

      _outtext("\n");
      sprintf(text,"Z ACCELERATION %0.3lf - ",start[i].accelz);
      _outtext(text);
      fflush( stdin );
      gets(in);
      if(!(iscntrl(in[0])))start[i].accelz=data_convert(in);

      _outtext("\n");
      sprintf(text,"FRICTION %0.4lf - ",start[i].friction);
      _outtext(text);
      fflush( stdin );
      gets(in);
      if(!(iscntrl(in[0])))start[i].friction=data_convert(in);

      _clearscreen(_GCLEARSCREEN);
      _outtext("                        ** SET START PARAMETERS **\n");

      ball[i]=start[i];
   }
}


double data_convert(char in[15])
{
   double data;
   sscanf(in, "%lf",&data);
   return(data);
}

void setscreen()                              /* get screen ready */
{
   _setvideomode(video);                /* vga 16 color */
   _rectangle(_GBORDER, 0,0,maxx+1,maxy+1);
   _setvieworg(xoffset,yoffset);

}


void reflectxlower(int num)                    /* reflect off of left side */
{
   double dist;
   double t;
   double store;

   if(ball[num].accelx !=0)
   {
      dist=ball[num].x-1;
      store=ball[num].velx*ball[num].velx-(2*ball[num].accelx*dist);
      if(store<.0003) store=.0003;
      ball[num].newvelx=sqrt(store);                      /* find final velocity */
      t=1-fabs((ball[num].newvelx+ball[num].velx)/ball[num].accelx);          /* rebound time */
      ball[num].newx=1+(ball[num].newvelx*t+(.5*ball[num].accelx*pow(t,2)));  /* new position */
      if(ball[num].newx<1) ball[num].newx=1;
      ball[num].newvelx=ball[num].newvelx+(ball[num].accelx*t);               /* new velocity */
   }

   else
      {                                         /* bounce with no accel */
      ball[num].newx=1+(-1*ball[num].newx);
      ball[num].newvelx=-1*ball[num].newvelx;
      }
   beep(10000,1000);

}

void reflectxupper(int num)                  /* bounce off the right side */
{
   double dist;
   double t;
   double store;

   if(ball[num].accelx !=0)
   {
      dist=xworld-ball[num].x;                               /* distance to wall */
      store=ball[num].velx*ball[num].velx+(2*ball[num].accelx*dist);
      if(store<.0003) store=.0003;
      ball[num].newvelx=sqrt(store);                       /* find final velocity */
      t=1-fabs((ball[num].newvelx-ball[num].velx)/ball[num].accelx);           /* rebound time */
      ball[num].newx=xworld-(ball[num].newvelx*t-(.5*ball[num].accelx*pow(t,2)));/* new position */
      if(ball[num].newx>xworld) ball[num].newx=xworld;
      ball[num].newvelx=-(ball[num].newvelx-(ball[num].accelx*t));             /* new velocity */
   }

   else
   {                                          /* bounce with no accel */
      ball[num].newx=xworld-(ball[num].newx-xworld);
      ball[num].newvelx=-ball[num].newvelx;
   }

     beep(10000,1000);

}

void reflectylower(int num)                 /* reflect off of bottom */
{
   double dist;
   double t;
   double store;

   if(ball[num].accely !=0)
   {
      dist=ball[num].y-1;
      store=ball[num].vely*ball[num].vely-(2*ball[num].accely*dist);
      if(store<.0003) store=.0003;
      ball[num].newvely=sqrt(store);                      /* find final velocity */
      t=1-fabs((ball[num].newvely+ball[num].vely)/ball[num].accely);          /* rebound time */
      ball[num].newy=1+(ball[num].newvely*t+(.5*ball[num].accely*pow(t,2)));  /* new position */
      if(ball[num].newy<1) ball[num].newy=1;
      ball[num].newvely=ball[num].newvely+(ball[num].accely*t);               /* new velocity */
   }

   else
   {                                         /* bounce with no accel */
      ball[num].newy=1+(-1*ball[num].newy);
      ball[num].newvely=-1*ball[num].newvely;
   }


   beep(10000,2000);
}

void reflectyupper(int num)                  /* bounce off the top */
{
   double dist;
   double t;
   double store;

   if(ball[num].accely !=0)
   {
      dist=yworld-ball[num].y;                               /* distance to wall */
      store=ball[num].vely*ball[num].vely+(2*ball[num].accely*dist);
      if(store<.0003) store=.0003;
      ball[num].newvely=sqrt(store);                       /* find final velocity */
      t=1-fabs((ball[num].newvely-ball[num].vely)/ball[num].accely);           /* rebound time */
      ball[num].newy=yworld-(ball[num].newvely*t-(.5*ball[num].accely*pow(t,2)));/* new position */
      if(ball[num].newy>yworld) ball[num].newy=yworld;
      ball[num].newvely=-(ball[num].newvely-(ball[num].accely*t));             /* new velocity */
   }

   else
   {                                          /* bounce with no accel */
      ball[num].newy=yworld-(ball[num].newy-yworld);
      ball[num].newvely=-ball[num].newvely;
   }

   beep(10000,2000);
}


void reflectzlower(int num)               /* bounce off the front */
{
   double dist;
   double t;
   double store;

   if(ball[num].accelz !=0)
   {
      dist=ball[num].z-1;
      store=ball[num].velz*ball[num].velz-(2*ball[num].accelz*dist);
      if(store<.0003) store=.0003;
      ball[num].newvelz=sqrt(store);                      /* find final velocity */
      t=1-fabs((ball[num].newvelz+ball[num].velz)/ball[num].accelz);          /* rebound time */
      ball[num].newz=1+(ball[num].newvelz*t+(.5*ball[num].accelz*pow(t,2)));  /* new position */
      if(ball[num].newz<1) ball[num].newz=1;
      ball[num].newvelz=ball[num].newvelz+(ball[num].accelz*t);               /* new velocity */
   }

   else
   {                                         /* bounce with no accel */
      ball[num].newz=1+(-1*ball[num].newz);
      ball[num].newvelz=-1*ball[num].newvelz;
   }

   beep(5000,5000);
   beep(5000,8000);
}

void reflectzupper(int num)                  /* bounce off the back */
{
   double dist;
   double t;
   double store;

   if(ball[num].accelz !=0)
   {
      dist=zworld-ball[num].z;                               /* distance to wall */
      store=ball[num].velz*ball[num].velz+(2*ball[num].accelz*dist);
      if(store<.0003) store=.0003;
      ball[num].newvelz=sqrt(store);                       /* find final velocity */
      t=1-fabs((ball[num].newvelz-ball[num].velz)/ball[num].accelz);           /* rebound time */
      ball[num].newz=zworld-(ball[num].newvelz*t-(.5*ball[num].accelz*pow(t,2)));/* new position */
      if(ball[num].newz>zworld) ball[num].newz=zworld;
      ball[num].newvelz=-(ball[num].newvelz-(ball[num].accelz*t));             /* new velocity */
   }

   else
   {                                          /* bounce with no accel */
      ball[num].newz=zworld-(ball[num].newz-zworld);
      ball[num].newvelz=-ball[num].newvelz;
   }

   beep(10000,700);

}



void movexyz(int num)                /* find new x,y and z velocity and position */
{
   ball[num].newx=ball[num].x+ball[num].velx+(.5*ball[num].accelx);
   ball[num].newy=ball[num].y+ball[num].vely+(.5*ball[num].accely);
   ball[num].newz=ball[num].z+ball[num].velz+(.5*ball[num].accelz);
   ball[num].newvelx=ball[num].velx+ball[num].accelx;
   ball[num].newvely=ball[num].vely+ball[num].accely;
   ball[num].newvelz=ball[num].velz+ball[num].accelz;
}



void slow(int num)                              /* friction affect? */
{
   ball[num].newvelx=ball[num].newvelx-(ball[num].newvelx * ball[num].friction);
   ball[num].newvely=ball[num].newvely-(ball[num].newvely * ball[num].friction);
   ball[num].newvelz=ball[num].newvelz-(ball[num].newvelz * ball[num].friction);

}



void beep( unsigned duration, unsigned frequency )
{
    int control;
    int i;

    if (frequency)   /* frequency must be > 0 to make sound */
    {
		     /* Prepare timer by sending 10111100 to port 43. */
	outp( 0x43, 0xb6 );

		   /* Divide input frequency by timer ticks per second and */
		   /* write (byte by byte) to timer. */

	frequency = (unsigned)(1193180L / frequency);
	outp( 0x42, (char)frequency );
	outp( 0x42, (char)(frequency >> 8) );

	control = inp( 0x61 );           /* Save speaker control byte. */
	outp( 0x61, control | 0x3 );     /* turn speaker on (bits 0 & 1) */

	for(i=duration;i>0;i--);              /* delay */
	if( frequency )
	   outp( 0x61, control );             /* turn speaker back off */
     }
}


int graphics_mode()                       /* make graphics happen */
{                                         /* go for the highest resolution */
   struct videoconfig screen;
   int mode=_VRES16COLOR;
   while( !_setvideomode( mode ))
      mode--;
   if(mode== _TEXTMONO)
   {
      mode--;
      while( !_setvideomode(mode))
	 mode--;
   }
   if(mode<_MRES4COLOR)
   {
      mode=0;
      _setvideomode(_DEFAULTMODE);
      printf("I CAN'T TURN THE GRAPHICS STUFF ON.  SEE YA!");
      exit(0);
   }

   _getvideoconfig(&screen);
   maxx = screen.numxpixels - 2;
   maxy = screen.numypixels - 2;
   _setvideomode(_DEFAULTMODE);
   return(mode);

}

void set_default(void)                 /* set up all default data */
{
   start[0].x=10000;
   start[0].y=10000;
   start[0].z=10000;
   start[0].velx=120;
   start[0].vely=123;
   start[0].velz=126;
   start[0].accelx=0;
   start[0].accely=0;
   start[0].accelz=0;
   start[0].friction=0;

   start[1].x=10000;
   start[1].y=10000;
   start[1].z=10000;
   start[1].velx=0;
   start[1].vely=0;
   start[1].velz=0;
   start[1].accelx=-1;
   start[1].accely=-2;
   start[1].accelz=0;
   start[1].friction=0;

   start[2].x=8000;
   start[2].y=15000;
   start[2].z=10;
   start[2].velx=0;
   start[2].vely=0;
   start[2].velz=90;
   start[2].accelx=0;
   start[2].accely=5;
   start[2].accelz=0;
   start[2].friction=0;

}


void project_world(void)               /* print 3D world to screen */
{
   ball[0].x=0;
   ball[0].y=0;
   ball[0].z=0;

   for(; ball[0].x < (xworld+1); ball[0].x +=1000)  /* vary x */
   {
      ball[0].y=yworld;
      point_convert(0);
      _moveto(ball[0].xscreen,ball[0].yscreen);
      ball[0].z=zworld;
      point_convert(0);
      _lineto(ball[0].xscreen,ball[0].yscreen);
      ball[0].y=0;
      point_convert(0);
      _lineto(ball[0].xscreen,ball[0].yscreen);
      ball[0].z=0;
      point_convert(0);
      _lineto(ball[0].xscreen,ball[0].yscreen);
      beep(2000,1000);

   }

   ball[0].x=0;
   ball[0].y=0;
   ball[0].z=0;

   for(; ball[0].z < (zworld+1); ball[0].z +=1000)       /* vary z */
   {
      point_convert(0);
      _moveto(ball[0].xscreen,ball[0].yscreen);
      ball[0].y=yworld;
      point_convert(0);
      _lineto(ball[0].xscreen,ball[0].yscreen);
      ball[0].x=xworld;
      point_convert(0);
      _lineto(ball[0].xscreen,ball[0].yscreen);
      ball[0].y=0;
      point_convert(0);
      _lineto(ball[0].xscreen,ball[0].yscreen);
      ball[0].x=0;
      point_convert(0);
      _lineto(ball[0].xscreen,ball[0].yscreen);
      beep(2000,2000);
   }

   ball[0].x=0;
   ball[0].y=0;
   ball[0].z=0;

   for(; ball[0].y < (yworld+1); ball[0].y +=1000)       /* vary y */
   {
      ball[0].x=xworld;
      point_convert(0);
      _moveto(ball[0].xscreen,ball[0].yscreen);
      ball[0].z=zworld;
      point_convert(0);
      _lineto(ball[0].xscreen,ball[0].yscreen);
      ball[0].x=0;
      point_convert(0);
      _lineto(ball[0].xscreen,ball[0].yscreen);
      ball[0].z=0;
      point_convert(0);
      _lineto(ball[0].xscreen,ball[0].yscreen);
      beep(2000,3000);
   }
}

void point_convert(int num)              /* convert xyz to screen xy */
{
   int totz;
   double x;
   double y;
   totz=ball[num].z+view+screendist;

   x=ball[num].x-xview;
   y=ball[num].y-yview;


   ball[num].yscreen=((y / totz) * view);
   ball[num].xscreen=((x / totz) * view);
}


char mainmenu(void)                     /* print main menu to screen */
{
   char selection;


   _clearscreen(_GCLEARSCREEN);
   _settextcolor(9);
   _outtext("\n\n*******************************************************************************\n\n");
   _settextcolor(2);
   _outtext("                 THE 3D BOUNCING BALL BY BRAD WHITSITT (V 1.0)\n");
   _outtext("                                 MAIN MENU\n\n\n\n\n");
   _outtext("  1. INSTRUCTIONS\n");
   _outtext("  2. SET START PARAMETERS\n");
   _outtext("  3. START\n");
   _outtext("  4. EXIT\n\n\n\n\n");
   _outtext("  INPUT YOUR OPTION:\n");
   _settextcolor(9);
   _outtext("*******************************************************************************");
   _settextcolor(2);
   _settextposition(19,22);
   fflush( stdin );
   do
       selection=getch();
   while(selection != '1' && selection !='2' && selection != '3' && selection != '4');

   return(selection);


}
   /* the end */
